;#######################################################################
;
; File:         mgraph.lsp  
; Description:  2d and 3d graphics routines for macsyma
;                 using X11 (translated from franz lisp version)
; Author:       Paul S. Wang
;
; (c) Copyright 1988,1989  Paul Wang, Kent State University 
;
;#######################################################################

;;     graphclose();
;;
;;    is used to close the x graphics window
;;
(in-package "MAXIMA")

(proclaim '(special $gfileonly pts overlay* $xangle $yangle $zangle hidden*
             dname* color* grid* param* $numer bgcolor* daxiscolor*
             gport $nolispsub name* dcapcolor* axiscolor* capcolor*
             dgrid* $xgraph doverlay* dxangle dyangle dzangle dhidden*
             dcolor* dbgcolor* line-style* fillcolor* dfillcolor* axis*
             daxis* *rr* *ss* isotropic* disotropic* *2d* *3d*
             local-obj-dir $color printer* dprinter* mag* dmag* hard*
             local-obj-dir hardfile*))  

;; location of executable xgraph
;; compiled by liszt with no -m option
(setq $xgraph "../XGP") ;;xgraph command bennett 

(setq $nolispsub nil)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; GRAPH - Top level routine called by the user,
;;	  this calls the particular graphing function needed.
;;
;; graph(funcz,[varx,vary],[lowx,highx,numpoints],[lowy,highy,numpoints]);
;;	z = funcz(varx,vary)
;; graph([fx,fy,fz],[varu,varv],[lowu,highu,upoints],[lowv,highv,vpoints]);
;;	x = fx(varu,varv)
;;	y = fy(varu,varv)
;;	z = fz(varu,varv)
;; graph(funcy,varx,[lowx,highx,numpoints]);  2d plot
;;	y = funcy(varx)
;; graph([fx,fy],var,[low,high,point]); 2d parametric plot
;;	x = fx(var)
;;	y = fy(var)
;;
(defmacro mlistp (x)
  `(and (listp ,x)
           (listp (car ,x))
           (eq (caar ,x) 'mlist)))
(defmacro meqp (x)
  `(and (listp ,x)
           (listp (car ,x))
           (eq (caar ,x) 'mequal)))
(defmacro mystring (atom)
  `(string-downcase (subseq (princ-to-string ,atom) 0)))

;; the pseudo term pair used to communicate with C-based xgraph

(setq $gfileonly nil) ;; if t no display just generate the file of points
(setq gport nil) ;; the graphics port flag
(setq $color nil)

;; setting macsyma flag for floating point evaluation
(defun $graph (&rest args)
  (prog (f fx fy fz u v lowu hiu nu lowv hiv nv op *2d* *3d* $numer
           fillcolor* color* bgcolor* axiscolor* hidden* overlay*
           $xangle $yangle $zangle axis* param* name* tmp nargs)

    (setq lowv 0 hiv 0 nv 0 $numer t nargs (length args))
    (setq f (car args) u (cadr args))
    (cond
      ((mlistp u) (setq *3d* t v (caddr u) u (cadr u)))
      (t (setq *2d* t)))
    (setq param* t)
    (cond
      ((mlistp f)
         (setq param* nil)
         (cond (*2d* (setq fx (cadr f) fy (caddr f)))
               (*3d* (setq fx (cadr f) fy (caddr f) fz (cadddr f)))))
      (*3d* (setq fx u fy v fz f))
      (*2d* (setq fx u fy f))
    )
    (setq tmp (cdr (meval (caddr args))))
    (setq lowu (car tmp) hiu (cadr tmp)  nu (caddr tmp) )
    (when *3d*
      (setq tmp (cdr (meval (cadddr args))))
      (setq lowv (car tmp) hiv (cadr tmp) nv (caddr tmp))
    )
    (cond
      ((and *2d* (= nargs 4)) (setq op (nth 3 args)))
      ((and *3d* (= nargs 5)) (setq op (nth 4 args))))
    (cond
      ((not (and (atom u) (atom v)))
       (merror "variables must be atomic, please check your args"))
      ((not (and (numberp lowu) (numberp lowv) (numberp hiu)
                 (numberp hiv)))
       (merror "plotting ranges must be numerical"))
      ((not (and (integerp nu) (integerp nv)))
       (merror "number of points must be integral")))
    (graphreset)
    (setq $numer t)
    (when op (graphswitch (cdr op)))
    (cond
      (*2d* (curve fx fy u lowu hiu nu))
      (*3d* (surf fx fy fz u lowu hiu nu v lowv hiv nv)))
    (return '$done)))

(defun $graphinit (color) (graphd1 color))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;surf-graphs a surface in 3D given
;	3 parameterized equations fx(u,v),fy(u,v),fz(u,v).The parameters
;	u and v will run from lowu to hiu and lowv to
;	hiv respectively, and nu * nv points will be generated
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; array of (x,y,z) points
(defun surf (fx fy fz u lowu hiu nu v lowv hiv nv)
  (prog (du dv np file)
    (if (> nu 1) (setq du (/ (float (- hiu lowu)) (1- nu)))
        (merror "must graph at least two points in each range"))
    (if (> nv 1) (setq dv (/ (float (- hiv lowv)) (1- nv)))
        (merror "must graph at least two points in each range"))
    (setq np (* nu nv))
    (setq pts  (make-array (list (+ 1 np) 3) :element-type 'float))
    (if (lispsub (list fx fy fz) u lowu nu v lowv nv du dv) nil
        (fsub (list '(mlist) fx fy fz) u lowu nu v lowv nv du dv))
    (setq file (dograph pts np name* nu nv))
    (return file)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; curve is the save as surf but for 2d graphs
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; array of (x,y) points
(defun curve (fx fy u lowu hiu nu)
  (prog (du np file)
    (if (> nu 1) (setq du (/ (float (- hiu lowu)) (1- nu)))
        (merror "must graph at least 2 points"))
    (setq np nu)
    (setq pts (make-array  (list (+ 1 np) 2) :element-type 'float))
    (fsub2 (list '(mlist) fx fy) u lowu nu du)
    (setq file (dograph pts np name*))
    (return file)))

;; fsub2: evaluation of points for 2d graphs

;;;
;;; evaluate the function and store
;;; in (nu x 2) array pts(m,n) 
;;;
(defun fsub2 (fxy u lowu nu du)
  (do ((m 0 (1+ m)) (a) (b) (uu) (evl)) ((= m nu) nil)
    (setq uu (+ lowu (* m du)))
    (setq evl `((mequal) ,u ,uu))
    (setq a (cdr (meval `(($ev) ,fxy ,evl))))
    (setq b (cadr a) a (car a))
    (cond
      ((and (floatp (setf (aref pts m 0) a)) (floatp (setf (aref pts m 1) b)))
       nil)
      (t (print "fsub2: point non numeric") (error "")))))

;; fsub finds point values (3d graph) by evaluating the functions
;; using macsyma

;;;
;;; evaluate the function and store
;;; in (nu *$ nv) x 3 array pts(m,n) 
;;;
(defun fsub (fxyz u lowu nu v lowv nv du dv)
  (do ((i 0 (1+ i)) (m 0) (a) (b) (c) (uu) (vv) (evl)) ((= i nu) nil)
    (setq uu (+ lowu (* i du)))
    (do ((j 0 (1+ j))) ((= j nv) nil)
      (setq vv (+ lowv (* j dv)))
      (setq evl
            `((mlist)
                 ((mequal) ,u
                  ,uu)
                 ((mequal) ,v
                  ,vv)))
      (setq a (cdr (meval `(($ev) ,fxyz ,evl))))
      (setq b (cadr a) c (caddr a) a (car a))
      (cond
        ((and (floatp (setf (aref pts m 0) a)) (floatp (setf (aref pts m 1) b))
              (floatp (setf (aref pts m 2) c)))
         nil)
        (t (print "fsub: value of point is not numeric") (terpri)
           (error "")))
      (setq m (1+ m)))))
;; try to convert all functions from vaxima rep to lisp rep
(defun trymtl (fl)
  (do ((f fl (cdr f)) (ans) (g)) ((null f) (nreverse ans))
    (setq g (catch 'gfoo (mtl (car f))))
    (when (null g) (return nil))
    (setq ans (cons g ans))))

(defun $regraph (dim op) ($graphfile dim op))

(defun $graphfile (dim op)
  (prog (*2d* *3d* $numer fillcolor* color* bgcolor* axiscolor* hidden*
              overlay* $xangle $yangle $zangle axis* param* name*)
    (if (equal dim 2) (setq *2d* t) (setq *3d* t))
    (graphreset)
    (when op (graphswitch (cdr op)))
    (dograph 0 0 name*)
    (return '$done)))
;; writes pts into a file and invoke C process to display
;; graphics
;; the C process works with X11 window and graphics

;; open output
;; call c graphing process to run in the background
;; debug  (print cmd) (terpri) 
(defun dograph (pts np name &optional nu nv)
  (prog (cmd output)
    (cond
      ($gfileonly (return nil))
      ((null gport)
       ;; sys::process implemented in process.c and proc.lsp
       (setq gport (sys::process $xgraph "w"))
       ;;;; later versions of akcl may have a run-process function
       ;;;;(setq gport (si::run-process $xgraph)) does not work
       (when (equal -1 gport)
         (merror "dograph: can't invoke ~m" $xgraph) (setq gport nil))
    ))
      (pr_points pts np nu nv))))

(defun pr_points (pts np &optional nu nv)
  (sys::prwrite
     (cond (*3d* (concatenate 'string
              "(DrawPrimitive surf (Surface " (princ-to-string nv)))
           (t (concatenate 'string "(DrawPrimitive fac (Polyline ")))
  )
  (do ((m 0 (1+ m)) (tab " ")) ((= m np) nil)
    (sys::prwrite (concatenate 'string  tab
       (princ-to-string (aref pts m 0)) tab
       (princ-to-string (aref pts m 1)) tab 
       (when *3d*
            (concatenate 'string (princ-to-string (aref pts m 2)) tab ))
    ))
  ) ;; end of do
  (sys::prwrite (concatenate 'string " ))"))
)

(defun $graphclose ()
  (sys::prwrite "(Quit)")
  (sys::prclose)
  (setq gport nil))
 


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun graphswitch (list)
  (do ((l list (cdr l)) (y nil)) ((null l))
    (setq y (car l))
    (cond
      ((not (meqp y)) (eprint "option" y))
      (t (setq y (mapcar #'stripdollar (cdr y)))
         (cond
           ((or (eq 'name (car y)) (eq 'file (car y)))
            (setq name* (mystring (cadr y))))
           ((eq 'printer (car y)) (setq printer* (mystring (cadr y))))
           ((eq 'mag (car y)) (setq mag* (cadr y)))
           ((eq 'background (car y))
            (cond
              ((setq bgcolor* (clr (cadr y))))
              (t (eprint "background color" (cadr y))
                 (setq bgcolor* dbgcolor*))))
           ((eq 'plotcolor (car y))
            (cond
              ((setq color* (clr (cadr y))))
              (t (eprint "color" (cadr y)) (setq color* dcolor*))))
           (*2d* (cond
                   ((eq 'isotropic (car y))
                    (setq isotropic*
                          (cond
                            ((eq (cadr y) 'true) t)
                            ((eq (cadr y) 'false) nil)
                            (t (eprint "isotropic option" (cadr y))
                               disotropic*))))
                   ((eq 'overlay (car y))
                    (setq overlay*
                          (cond
                            ((eq (cadr y) 'true) t)
                            ((eq (cadr y) 'false) nil)
                            (t (eprint "overlay value" (cadr y))
                               'false))))
                   ((eq 'grid (car y))
                    (when (setq grid* (if (eq (cadr y) 'true) t nil))))
                   ((eq 'axiscolor (car y))
                    (cond
                      ((setq axiscolor* (clr (cadr y))))
                      (t (eprint "axiscolor" (cadr y))
                         (setq axiscolor* daxiscolor*))))
                   ((eq 'capcolor (car y))
                    (cond
                      ((setq capcolor* (clr (cadr y))))
                      (t (eprint "capcolor" (cadr y))
                         (setq capcolor* dcapcolor*))))
                   (t (eprint "2D option" (car y)))))
           (*3d* (cond
                   ((eq 'fillcolor (car y))
                    (cond
                      ((setq fillcolor* (clr (cadr y))))
                      (t (eprint "fillcolor" (cadr y))
                         (setq fillcolor* dfillcolor*))))
                   ((eq 'hidden (car y))
                    (setq hidden*
                          (cond
                            ((eq (cadr y) 'true) t)
                            ((eq (cadr y) 'false) nil)
                            (t (eprint "hidden option" (cadr y))
                               dhidden*))))
                   ((eq 'xangle (car y))
                    (setq $xangle
                          (cond
                            ((atom (cadr y)) (cadr y))
                            (t (eprint "xangle option") dxangle))))
                   ((eq 'yangle (car y))
                    (setq $yangle
                          (cond
                            ((atom (cadr y)) (cadr y))
                            (t (eprint "yangle option") dyangle))))
                   ((eq 'zangle (car y))
                    (setq $zangle
                          (cond
                            ((atom (cadr y)) (cadr y))
                            (t (eprint "zangle option") dzangle))))
                   (t (eprint "3D option" (car y)))))
           (t (merror "~m unknown graph option" y)))))))

(defun eprint (s e) (print (list e '|INVALID | s))) 


;; isotropic = true => same scale for x and y
;; preserving shape
;; if nil plot will fill graphics window
;; with the x and y scaling treated independently
;; can also be paintjet
;; no axes when in 3D
;; if overlay is true plots are overlaid
;; no hidden line removal
;; no lines
;; fill color for polygons is cyan/white
;; background color is cyan/white 
;; default color is magenta (if b/w, default black)
(defun graphd1 (a)
  (if (or (eq a '$true) (eq a t)) (setq a t) (setq a nil))
  (setq dxangle 40.0 dyangle 50.0 dzangle 60.0)
  (setq disotropic* t dname* "surface" dprinter* "laserjet" dmag*
        3 hard* nil daxis* nil doverlay* nil dhidden* nil dgrid* nil
        daxiscolor* 'black dcapcolor* 'black dfillcolor*
        (if a 'cyan 'white) dbgcolor* dfillcolor* line-style* 0 dcolor*
        (if a 'magenta 'black))
  'graphd1)



;; these are 2D things
;; solid line to start with
;; for 2d only
;; overlays to start with
;; these are 3D things
;; these are both 2D and 3D things  
(defun graphreset ()
  (when (null (si:getenv "DISPLAY"))
    (print "Please set the environment variable DISPLAY")
    (terpri)
    (print "to hostname:0.0")
    (terpri)
    (print "before proceeding further with graph")
    (error ""))
  (cond
    (*2d* (setq capcolor* dcapcolor* line-style* 0 grid* dgrid*
                overlay* doverlay* isotropic* disotropic* axiscolor*
                daxiscolor*))
    (*3d* (setq hidden* dhidden* $xangle dxangle $yangle dyangle
                $zangle dzangle)))
  (setq bgcolor* dbgcolor* fillcolor* dfillcolor* color* dcolor*
        printer* dprinter* mag* dmag* name* dname*))


(defun clr (a) (mystring a))

(graphd1 '$false)
;;;;    Usage progname [ options ] default value -2d
;;;;    |false| -3d |false| -x xangle 30 -y yangle 40 -z zangle 70 -fg
;;;;    color black -bg color white -fill color black -h do hidden line
;;;;    remove -grid -1 (no grid) false -file name ellipse/torus)

;; to speed up function evaluation
;; use lisp eval instead of vaxima meval in most cases

;; mtl converts a vaxima expression into a lisp expression
;; when this is not possible nil is returned

(defun mtl (exp)
  (prog (op n d)
    (when (atom exp) (return exp))
    (setq op (get (caar exp) 'lop))
    (cond
      ((null op) (throw 'gfoo nil))
      ((eq op '/)
       (setq n (mapcar #'mtl (cdr exp)))
       (setq d (cadr n) n (car n))
       (cond
         ((integerp n) (setq n (float n)))
         ((integerp d) (setq d (float d))))
       (return (cons op (list n d)))))
    (return (cons op (mapcar #'mtl (cdr exp))))))


(mapc #'(lambda (a) (setf (get (car a) 'lop) (cadr a)))
      '((mexpt expt) (mplus +)
        (mminus -) (mdifference -)
        (mtimes *) (mquotient /)
        (%sin sin) (%cos cos) (%tan tan) 
	(%acos acos) (%asin asin) (%atan atan)
	(%log log)
	(%acosh acosh) (%asinh asinh) (%atanh atanh)
	(%sinh sinh) (%cosh cosh) (%tanh tanh) 
        (rat /)))

;; try lisp level substitution 
(proclaim '(special *rr* *ss*))
;;;
;;; evaluate the function and store
;;; in (nu X nv) x 3 array pts(m,n) 
;;;
(defun lispsub (fxyz u lowu nu v lowv nv du dv)
  (prog (foo* uu vv)
    (when $nolispsub (return nil))
    (setq foo* t)
    (setq fxyz (trymtl fxyz))
    (when (null fxyz) (return nil))
    (setq fxyz (sublis (list (cons u '*rr*) (cons v '*ss*)) fxyz))
    (do ((i 0 (1+ i)) (m 0)) ((= i nu) nil)
      (setq uu (+ lowu (* i du)))
      (do ((j 0 (1+ j))) ((= j nv) nil)
        (setq vv (+ lowv (* j dv)))
        (setq *rr* uu)
        (setq *ss* vv)
	;; errset not yet implemented for CL
        (setf (aref pts m 0) (float (eval (car   fxyz))))
        (setf (aref pts m 1) (float (eval (cadr  fxyz))))
        (setf (aref pts m 2) (float (eval (caddr fxyz))))
        (setq m (1+ m)))
      )
    (return t)
))

(when (not (functionp 'process))
      ;;;(sys::faslink "proc.o" "process.o -lc")
      ;; it seems that -lc is not needed
      (sys::faslink (concatenate 'string  local-obj-dir "/proc.o")
                    (concatenate 'string  local-obj-dir "/process.o")
      )
)

